from decimal import Decimal
from operator import (
    add,
    eq,
    floordiv,
    ge,
    gt,
    iand,
    inv,
    ior,
    is_,
    is_not,
    le,
    lshift,
    lt,
    mod,
    ne,
    neg,
    not_,
    pos,
    rshift,
    sub,
    truediv,
)
from types import NoneType
from typing import Any, TypeGuard, TypeVar, cast, overload

from django.db.models import BooleanField, Field, Model, Q
from django.db.models.expressions import (
    BaseExpression,
    Case,
    Combinable,
    F,
    Value,
    When,
)
from django.db.models.functions import Floor
from django.db.models.lookups import (
    Exact,
    GreaterThan,
    GreaterThanOrEqual,
    In,
    IsNull,
    LessThan,
    LessThanOrEqual,
)
from django.db.models.options import Options

from django_xformula.db.lookups import Pure

__all__ = [
    "BidirectionalOperator",
]


T = TypeVar("T")


class BidirectionalOperator:
    class QEnum:

        EMPTY = Q()

        # Pointer; will not belong to any query.
        NEGATED = ~Q(Exact(Value(False), Value(False)))

    class Optimization:

        LITERAL_TYPES = (
            NoneType,
            bool,
            int,
            float,
            complex,
            str,
        )

        EXCLUDED_TYPES: dict[str, tuple[type, ...]] = dict(
            # Floating-point arithmetic implementations may vary
            # between DBs / runtimes.
            # For consistency of results, get operations done, in DBs.
            arithmetic=(float, Decimal),
            bitwise=(float, Decimal),
            compare=(float, Decimal),
            logical=(float, Decimal),
        )

    @classmethod
    def is_q(
        cls,
        py_value: Any,
    ) -> TypeGuard[Q]:
        return isinstance(py_value, Q)

    @classmethod
    @overload
    def is_q_or_combinable(
        cls,
        q: Q,
    ) -> TypeGuard[Q]:
        ...

    @classmethod
    @overload
    def is_q_or_combinable(
        cls,
        combinable: Combinable,
    ) -> TypeGuard[Combinable]:
        ...

    @classmethod
    def is_q_or_combinable(
        cls,
        py_value,
    ):
        return cls.is_q(py_value) or cls.is_combinable(py_value)

    @classmethod
    def is_expression(
        cls,
        py_value: Any,
    ) -> TypeGuard[BaseExpression]:
        return isinstance(py_value, BaseExpression)

    @classmethod
    def is_combinable(
        cls,
        py_value: Any,
    ) -> TypeGuard[Combinable]:
        return isinstance(py_value, Combinable)

    @classmethod
    def is_f(
        cls,
        py_value: Any,
    ) -> TypeGuard[F]:
        return isinstance(py_value, F)

    @classmethod
    def is_value(
        cls,
        py_value: Any,
    ) -> TypeGuard[Value]:
        return isinstance(py_value, Value)

    @classmethod
    def is_model_instance(
        cls,
        py_value: Any,
    ) -> TypeGuard[Model]:
        return isinstance(py_value, Model)

    @classmethod
    def is_field(
        cls,
        py_value: Any,
    ) -> TypeGuard[Field]:
        return isinstance(py_value, Field)

    @classmethod
    def is_constant(
        cls,
        py_value: Any,
    ) -> TypeGuard[Any]:
        py_value = cls.ensure_if_py_value(py_value)
        return isinstance(py_value, cls.Optimization.LITERAL_TYPES)

    @classmethod
    def any_q(
        cls,
        *py_values: Any,
    ) -> bool:
        return any(map(cls.is_q, py_values))

    @classmethod
    def any_combinable(
        cls,
        *py_values: Any,
    ) -> bool:
        return any(map(cls.is_combinable, py_values))

    @classmethod
    def any_constant(
        cls,
        *maybe_combinables: Any,
    ) -> bool:
        return any(map(cls.is_constant, maybe_combinables))

    @classmethod
    def any_q_or_combinable(
        cls,
        *py_values: Any,
    ) -> bool:
        return any(
            map(
                lambda py_value: (cls.is_q(py_value) or cls.is_combinable(py_value)),
                py_values,
            ),
        )

    @classmethod
    @overload
    def ensure_if_combinable(
        cls,
        combinable: Combinable,
    ) -> Combinable:
        ...

    @classmethod
    @overload
    def ensure_if_combinable(
        cls,
        model_instance: Model,
    ) -> Any:
        ...

    @classmethod
    @overload
    def ensure_if_combinable(
        cls,
        py_value: T,
    ) -> T:
        ...

    @classmethod
    def ensure_if_combinable(
        cls,
        maybe_combinable,
    ):
        if isinstance(maybe_combinable, (list, set, tuple)):
            container_type = type(maybe_combinable)
            return container_type(map(cls.ensure_if_combinable, maybe_combinable))

        if cls.is_combinable(maybe_combinable):
            combinable = cast(Combinable, maybe_combinable)
            return combinable

        if cls.is_model_instance(maybe_combinable):
            model_instance = cast(Model, maybe_combinable)
            model_options = cast(Options, getattr(model_instance.__class__, "_meta"))
            if model_options.abstract or model_options.pk is None:
                return model_instance
            pk_attname = cls.ensure_field_attname(model_options.pk)
            pk = getattr(model_instance, pk_attname)
            return Value(pk)

        if not cls.is_q(maybe_combinable) and not cls.is_expression(maybe_combinable):
            value = Value(maybe_combinable)
            return value

        return maybe_combinable

    @classmethod
    @overload
    def ensure_if_py_value(
        cls,
        value: Value,
    ) -> Any:
        ...

    @classmethod
    @overload
    def ensure_if_py_value(
        cls,
        py_value: T,
    ) -> T:
        ...

    @classmethod
    def ensure_if_py_value(
        cls,
        maybe_value,
    ):
        if cls.is_value(maybe_value):
            value = cast(Value, maybe_value)
            return value.value
        return maybe_value

    @classmethod
    @overload
    def ensure_field_attname(
        cls,
        field: Field,
    ) -> str:
        ...

    @classmethod
    @overload
    def ensure_field_attname(
        cls,
        py_value: T,
    ) -> T:
        ...

    @classmethod
    def ensure_field_attname(
        cls,
        py_value,
    ):
        if cls.is_field(py_value):
            field = cast(Field, py_value)
            return field.attname

        return py_value

    @classmethod
    def split_qs(
        cls,
        *py_values: Any,
    ) -> tuple[list[Q], list[Combinable]]:
        qs: list[Q] = []
        combinables: list[Combinable] = []

        for py_value in py_values:
            if cls.is_q(py_value):
                q = cast(Q, py_value)
                qs.append(q)
            else:
                combinable = cls.ensure_if_combinable(py_value)
                combinables.append(combinable)

        return qs, combinables

    @classmethod
    def split_values(
        cls,
        *maybe_combinables: Any,
    ) -> tuple[list[Value], list[Any]]:
        values: list[Value] = []
        py_values: list[Any] = []

        for maybe_combinable in maybe_combinables:
            maybe_combinable = cls.ensure_if_combinable(maybe_combinable)

            if cls.is_value(maybe_combinable):
                value = cast(Value, maybe_combinable)
                values.append(value)
            else:
                py_value = cls.ensure_if_py_value(maybe_combinable)
                py_values.append(py_value)

        return values, py_values

    @classmethod
    def is_falsy_combinable(
        cls,
        maybe_combinable: Any,
    ) -> bool:
        py_value = cls.ensure_if_py_value(maybe_combinable)
        return not py_value

    @classmethod
    def should_optimize_for(
        cls,
        operation_name: str,
        *maybe_combinables: Any,
    ) -> bool:
        if operation_name not in cls.Optimization.EXCLUDED_TYPES:
            return True
        types = cls.Optimization.EXCLUDED_TYPES[operation_name]
        return all(
            map(
                lambda py_value: not isinstance(py_value, types),
                map(
                    cls.ensure_if_py_value,
                    maybe_combinables,
                ),
            ),
        )

    @classmethod
    def q_test(
        cls,
        maybe_combinable: Any,
    ) -> Q:
        if cls.is_q(maybe_combinable):
            q = cast(Q, maybe_combinable)
            return q

        combinable = cls.ensure_if_combinable(maybe_combinable)

        if cls.is_falsy_combinable(combinable):
            return cls.QEnum.NEGATED

        elif cls.is_value(combinable):
            return cls.QEnum.EMPTY

        return Q(
            Case(
                When(
                    Pure(combinable),
                    then=Value(True),
                ),
                default=Value(False),
                output_field=BooleanField(),
            ),
        )

    @classmethod
    def q_and(
        cls,
        q1: Q,
        q2: Q,
    ) -> Q:
        if not q1.children and not q2.children:
            return cls.QEnum.EMPTY

        if q1.children and not q2.children:
            return q1

        if not q1.children and q2.children:
            return q2

        if q1 is cls.QEnum.NEGATED and q2 is cls.QEnum.NEGATED:
            return cls.QEnum.NEGATED

        if q1 is cls.QEnum.NEGATED and q2 is not cls.QEnum.NEGATED:
            return ~q2

        if q1 is not cls.QEnum.NEGATED and q2 is cls.QEnum.NEGATED:
            return ~q1

        return q1 & q2

    @classmethod
    def q_xor(
        cls,
        q1: Q,
        q2: Q,
    ) -> Q:
        if not q1.children and not q2.children:
            return cls.QEnum.EMPTY

        if q1.children and not q2.children:
            return ~q1

        if not q1.children and q2.children:
            return ~q2

        if q1 is cls.QEnum.NEGATED and q2 is cls.QEnum.NEGATED:
            return cls.QEnum.NEGATED

        if q1 is cls.QEnum.NEGATED and q2 is not cls.QEnum.NEGATED:
            return q2

        if q1 is not cls.QEnum.NEGATED and q2 is cls.QEnum.NEGATED:
            return q1

        return q1 ^ q2

    @classmethod
    def q_or(
        cls,
        q1: Q,
        q2: Q,
    ) -> Q:
        if not q1.children and not q2.children:
            return cls.QEnum.EMPTY

        if q1.children and not q2.children:
            return q1

        if not q1.children and q2.children:
            return q2

        if q1 is cls.QEnum.NEGATED and q2 is cls.QEnum.NEGATED:
            return cls.QEnum.NEGATED

        if q1 is cls.QEnum.NEGATED and q2 is not cls.QEnum.NEGATED:
            return q2

        if q1 is not cls.QEnum.NEGATED and q2 is cls.QEnum.NEGATED:
            return q1

        return q1 | q2

    @classmethod
    def pos(
        cls,
        operand: Any,
    ) -> Any:
        if not cls.is_combinable(operand):
            return pos(operand)

        operand = cls.ensure_if_combinable(operand)
        return operand

    @classmethod
    def neg(
        cls,
        operand: Any,
    ) -> Any:
        if not cls.is_combinable(operand):
            return neg(operand)

        operand = cls.ensure_if_combinable(operand)

        if cls.should_optimize_for("arithmetic", operand):

            if cls.is_value(operand):
                operand = cast(Value, operand)
                return Value(neg(operand.value))

        return neg(operand)

    @classmethod
    def not_(
        cls,
        operand: Any,
    ) -> Any:
        if not cls.is_q_or_combinable(operand):
            return not_(operand)

        operand = cls.ensure_if_combinable(operand)

        if cls.is_q(operand):
            q = cast(Q, operand)

            if not q.children:
                if not q.negated:
                    return cls.QEnum.NEGATED
                else:
                    return cls.QEnum.EMPTY
            return ~q

        if not cls.is_value(operand):
            return ~cls.q_test(operand)

        if cls.should_optimize_for("logical", operand):

            if cls.is_value(operand):
                operand = cast(Value, operand)
                return Value(not_(operand.value))

        return not_(operand)

    @classmethod
    def inv(
        cls,
        operand: Any,
    ) -> Any:
        if not cls.is_q_or_combinable(operand):
            return inv(operand)

        operand = cls.ensure_if_combinable(operand)

        if cls.should_optimize_for("bitwise", operand):

            if cls.is_value(operand):
                operand = cast(Value, operand)
                return Value(inv(operand.value))

        if cls.is_q(operand):
            q = cast(Q, operand)
            return ~q

        return inv(operand)

    @classmethod
    def pow(
        cls,
        lhs: Any,
        rhs: Any,
    ) -> Any:
        if not cls.any_combinable(lhs, rhs):
            return pow(lhs, rhs)

        lhs = cls.ensure_if_combinable(lhs)
        rhs = cls.ensure_if_combinable(rhs)

        if cls.should_optimize_for("arithmetic", lhs, rhs):

            if cls.is_value(lhs) and cls.is_value(rhs):
                lhs = cast(Value, lhs)
                rhs = cast(Value, rhs)
                return Value(pow(lhs.value, rhs.value))

        return pow(lhs, rhs)

    @classmethod
    def mul(
        cls,
        lhs: Any,
        rhs: Any,
    ) -> Any:
        if not cls.any_combinable(lhs, rhs):
            return lhs * rhs

        lhs = cls.ensure_if_combinable(lhs)
        rhs = cls.ensure_if_combinable(rhs)

        if cls.should_optimize_for("arithmetic", lhs, rhs):

            if cls.is_value(lhs) and cls.is_value(rhs):
                lhs = cast(Value, lhs)
                rhs = cast(Value, rhs)
                return Value(lhs.value * rhs.value)

        return lhs * rhs

    @classmethod
    def truediv(
        cls,
        lhs: Any,
        rhs: Any,
    ) -> Any:
        if not cls.any_combinable(lhs, rhs):
            return truediv(lhs, rhs)

        lhs = cls.ensure_if_combinable(lhs)
        rhs = cls.ensure_if_combinable(rhs)

        if cls.should_optimize_for("arithmetic", lhs, rhs):

            if cls.is_value(lhs) and cls.is_value(rhs):
                lhs = cast(Value, lhs)
                rhs = cast(Value, rhs)
                return Value(truediv(lhs.value, rhs.value))

        return truediv(lhs, rhs)

    @classmethod
    def floordiv(
        cls,
        lhs: Any,
        rhs: Any,
    ) -> Any:
        if not cls.any_combinable(lhs, rhs):
            return floordiv(lhs, rhs)

        lhs = cls.ensure_if_combinable(lhs)
        rhs = cls.ensure_if_combinable(rhs)

        if cls.should_optimize_for("arithmetic", lhs, rhs):

            if cls.is_value(lhs) and cls.is_value(rhs):
                lhs = cast(Value, lhs)
                rhs = cast(Value, rhs)
                return Value(floordiv(lhs.value, rhs.value))

        return Floor(lhs / rhs)

    @classmethod
    def mod(
        cls,
        lhs: Any,
        rhs: Any,
    ) -> Any:
        if not cls.any_combinable(lhs, rhs):
            return mod(lhs, rhs)

        lhs = cls.ensure_if_combinable(lhs)
        rhs = cls.ensure_if_combinable(rhs)

        if cls.should_optimize_for("arithmetic", lhs, rhs):

            if cls.is_value(lhs) and cls.is_value(rhs):
                lhs = cast(Value, lhs)
                rhs = cast(Value, rhs)
                return Value(mod(lhs.value, rhs.value))

        return mod(lhs, rhs)

    @classmethod
    def add(
        cls,
        lhs: Any,
        rhs: Any,
    ) -> Any:
        if not cls.any_combinable(lhs, rhs):
            return add(lhs, rhs)

        lhs = cls.ensure_if_combinable(lhs)
        rhs = cls.ensure_if_combinable(rhs)

        if cls.should_optimize_for("arithmetic", lhs, rhs):

            if cls.is_value(lhs) and cls.is_value(rhs):
                lhs = cast(Value, lhs)
                rhs = cast(Value, rhs)
                return Value(add(lhs.value, rhs.value))

        return add(lhs, rhs)

    @classmethod
    def sub(
        cls,
        lhs: Any,
        rhs: Any,
    ) -> Any:
        if not cls.any_combinable(lhs, rhs):
            return sub(lhs, rhs)

        lhs = cls.ensure_if_combinable(lhs)
        rhs = cls.ensure_if_combinable(rhs)

        if cls.should_optimize_for("arithmetic", lhs, rhs):

            if cls.is_value(lhs) and cls.is_value(rhs):
                lhs = cast(Value, lhs)
                rhs = cast(Value, rhs)
                return Value(sub(lhs.value, rhs.value))

        return sub(lhs, rhs)

    @classmethod
    def lshift(
        cls,
        lhs: Any,
        rhs: Any,
    ) -> Any:
        if not cls.any_combinable(lhs, rhs):
            return lshift(lhs, rhs)

        lhs = cls.ensure_if_combinable(lhs)
        rhs = cls.ensure_if_combinable(rhs)

        return lhs.bitleftshift(rhs)

    @classmethod
    def rshift(
        cls,
        lhs: Any,
        rhs: Any,
    ) -> Any:
        if not cls.any_combinable(lhs, rhs):
            return rshift(lhs, rhs)

        lhs = cls.ensure_if_combinable(lhs)
        rhs = cls.ensure_if_combinable(rhs)

        return lhs.bitrightshift(rhs)

    @classmethod
    def and_(
        cls,
        lhs: Any,
        rhs: Any,
    ) -> Any:
        if not cls.any_q_or_combinable(lhs, rhs):
            return lhs and rhs

        qs, combinables = cls.split_qs(lhs, rhs)

        if not qs and combinables:
            return cls.q_test(combinables[0]) & cls.q_test(combinables[1])

        if qs and not combinables:
            return cls.q_and(qs[0], qs[1])

        if qs and combinables:
            q1 = qs[0]

            if cls.is_falsy_combinable(combinables[0]):
                return ~q1

            elif cls.is_value(combinables[0]):
                return q1

            else:
                q2 = cls.q_test(combinables[0])
                return cls.q_and(q1, q2)

        raise NotImplementedError(
            f"{cls.__qualname__}.and_ is not implemented for: " f" {[lhs, rhs]!r}"
        )

    @classmethod
    def iand(
        cls,
        lhs: Any,
        rhs: Any,
    ) -> Any:
        if not cls.any_q_or_combinable(lhs, rhs):
            return iand(lhs, rhs)

        qs, combinables = cls.split_qs(lhs, rhs)

        if not qs and combinables:
            return combinables[0].bitand(combinables[1])  # type: ignore[arg-type]

        if qs and not combinables:
            return cls.q_and(qs[0], qs[1])

        if qs and combinables:
            q1 = qs[0]

            if cls.is_falsy_combinable(combinables[0]):
                return ~q1

            elif cls.is_value(combinables[0]):
                return q1

            else:
                q2 = cls.q_test(combinables[0])
                return cls.q_and(q1, q2)

        raise NotImplementedError(
            f"{cls.__qualname__}.iand is not implemented for: " f" {[lhs, rhs]!r}"
        )

    @classmethod
    def xor(
        cls,
        lhs: Any,
        rhs: Any,
    ) -> Any:
        if not cls.any_q_or_combinable(lhs, rhs):
            return lhs ^ rhs

        qs, combinables = cls.split_qs(lhs, rhs)

        if not qs and combinables:
            return combinables[0].bitxor(combinables[1])  # type: ignore[arg-type]

        if qs and not combinables:
            return cls.q_xor(qs[0], qs[1])

        if qs and combinables:
            q1 = qs[0]

            if cls.is_falsy_combinable(combinables[0]):
                return q1

            elif cls.is_value(combinables[0]):
                return q1

            else:
                q2 = cls.q_test(combinables[0])
                return cls.q_xor(q1, q2)

        raise NotImplementedError(
            f"{cls.__qualname__}.xor is not implemented for: {[lhs, rhs]!r}",
        )

    @classmethod
    def or_(
        cls,
        lhs: Any,
        rhs: Any,
    ) -> Any:
        if not cls.any_q_or_combinable(lhs, rhs):
            return lhs or rhs

        qs, combinables = cls.split_qs(lhs, rhs)

        if not qs and combinables:
            return cls.q_test(combinables[0]) | cls.q_test(combinables[1])

        if qs and not combinables:
            return cls.q_or(qs[0], qs[1])

        if qs and combinables:
            q1 = qs[0]

            if cls.is_falsy_combinable(combinables[0]):
                return q1

            elif cls.is_value(combinables[0]):
                return q1

            else:
                q2 = cls.q_test(combinables[0])
                return cls.q_or(q1, q2)

        raise NotImplementedError(
            f"{cls.__qualname__}.or_ is not implemented for: " f" {[lhs, rhs]!r}"
        )

    @classmethod
    def ior(
        cls,
        lhs: Any,
        rhs: Any,
    ) -> Any:
        if not cls.any_q_or_combinable(lhs, rhs):
            return ior(lhs, rhs)

        qs, combinables = cls.split_qs(lhs, rhs)

        if not qs and combinables:
            return combinables[0].bitor(
                cast(int, combinables[1]),
            )

        if qs and not combinables:
            return cls.q_or(qs[0], qs[1])

        if qs and combinables:
            q1 = qs[0]

            if cls.is_falsy_combinable(combinables[0]):
                return q1

            elif cls.is_value(combinables[0]):
                return q1

            else:
                q2 = cls.q_test(combinables[0])
                return cls.q_or(q1, q2)

        raise NotImplementedError(
            f"{cls.__qualname__}.ior is not implemented for: " f" {[lhs, rhs]!r}"
        )

    @classmethod
    def nin(
        cls,
        lhs: Any,
        rhs: Any,
    ) -> Any:
        if not cls.any_combinable(lhs, rhs):
            return lhs not in rhs

        lhs = cls.ensure_if_combinable(lhs)
        rhs = cls.ensure_if_combinable(rhs)

        return ~Q(In(lhs, rhs))

    @classmethod
    def in_(
        cls,
        lhs: Any,
        rhs: Any,
    ) -> Any:
        if not cls.any_combinable(lhs, rhs):
            return lhs in rhs

        lhs = cls.ensure_if_combinable(lhs)
        rhs = cls.ensure_if_combinable(rhs)

        return Q(In(lhs, rhs))

    @classmethod
    def is_not(
        cls,
        lhs: Any,
        rhs: Any,
    ) -> Any:
        if not cls.any_combinable(lhs, rhs):

            if cls.any_constant(lhs, rhs):
                return ne(lhs, rhs)

            return is_not(lhs, rhs)

        lhs = cls.ensure_if_combinable(lhs)
        rhs = cls.ensure_if_combinable(rhs)

        values, py_values = cls.split_values(lhs, rhs)

        if cls.should_optimize_for("compare", lhs, rhs):

            if values and not py_values:

                if values[0].value is None and values[1].value is None:
                    return Value(False)

                if values[0].value == values[1].value:
                    return Value(False)

                else:
                    return Value(True)

            elif py_values and not values:
                return ~Q(Exact(lhs, rhs))

        value = values[0]
        py_value = py_values[0]

        if value.value is None:
            return ~Q(IsNull(py_value, True))

        return ~Q(Exact(lhs, rhs))

    @classmethod
    def is_(
        cls,
        lhs: Any,
        rhs: Any,
    ) -> Any:
        if not cls.any_combinable(lhs, rhs):

            if cls.any_constant(lhs, rhs):
                return eq(lhs, rhs)

            return is_(lhs, rhs)

        lhs = cls.ensure_if_combinable(lhs)
        rhs = cls.ensure_if_combinable(rhs)

        values, py_values = cls.split_values(lhs, rhs)

        if cls.should_optimize_for("compare", lhs, rhs):

            if values and not py_values:

                if values[0].value is None and values[1].value is None:
                    return Value(True)

                if values[0].value == values[1].value:
                    return Value(True)
                else:
                    return Value(False)

            elif py_values and not values:
                return Q(Exact(lhs, rhs))

        value = values[0]
        py_value = py_values[0]

        if value.value is None:
            return Q(IsNull(py_value, True))

        return Q(Exact(lhs, rhs))

    @classmethod
    def gt(
        cls,
        lhs: Any,
        rhs: Any,
    ) -> Any:
        if not cls.any_combinable(lhs, rhs):
            return gt(lhs, rhs)

        lhs = cls.ensure_if_combinable(lhs)
        rhs = cls.ensure_if_combinable(rhs)

        if cls.should_optimize_for("compare", lhs, rhs):
            values, py_values = cls.split_values(lhs, rhs)

            if values and not py_values:

                if values[0].value is None or values[1].value is None:
                    return Value(False)

                if values[0].value > values[1].value:
                    return Value(True)

                else:
                    return Value(False)

            elif values and py_values and values[0].value is None:
                return Value(False)

        return Q(GreaterThan(lhs, rhs))

    @classmethod
    def ge(
        cls,
        lhs: Any,
        rhs: Any,
    ) -> Any:
        if not cls.any_combinable(lhs, rhs):
            return ge(lhs, rhs)

        lhs = cls.ensure_if_combinable(lhs)
        rhs = cls.ensure_if_combinable(rhs)

        if cls.should_optimize_for("compare", lhs, rhs):
            values, py_values = cls.split_values(lhs, rhs)

            if values and not py_values:

                if values[0].value is None or values[1].value is None:
                    return Value(False)

                if values[0].value >= values[1].value:
                    return Value(True)

                else:
                    return Value(False)

            elif values and py_values and values[0].value is None:
                return Value(False)

        return Q(GreaterThanOrEqual(lhs, rhs))

    @classmethod
    def le(
        cls,
        lhs: Any,
        rhs: Any,
    ) -> Any:
        if not cls.any_combinable(lhs, rhs):
            return le(lhs, rhs)

        lhs = cls.ensure_if_combinable(lhs)
        rhs = cls.ensure_if_combinable(rhs)

        if cls.should_optimize_for("compare", lhs, rhs):
            values, py_values = cls.split_values(lhs, rhs)

            if values and not py_values:

                if values[0].value is None or values[1].value is None:
                    return Value(False)

                if values[0].value <= values[1].value:
                    return Value(True)

                else:
                    return Value(False)

            elif values and py_values and values[0].value is None:
                return Value(False)

        return Q(LessThanOrEqual(lhs, rhs))

    @classmethod
    def lt(
        cls,
        lhs: Any,
        rhs: Any,
    ) -> Any:
        if not cls.any_combinable(lhs, rhs):
            return lt(lhs, rhs)

        lhs = cls.ensure_if_combinable(lhs)
        rhs = cls.ensure_if_combinable(rhs)

        if cls.should_optimize_for("compare", lhs, rhs):
            values, py_values = cls.split_values(lhs, rhs)

            if values and not py_values:

                if values[0].value is None or values[1].value is None:
                    return Value(False)

                if values[0].value < values[1].value:
                    return Value(True)

                else:
                    return Value(False)

            elif values and py_values and values[0].value is None:
                return Value(False)

        return Q(LessThan(lhs, rhs))

    @classmethod
    def ne(
        cls,
        lhs: Any,
        rhs: Any,
    ) -> Any:
        if not cls.any_combinable(lhs, rhs):
            return ne(lhs, rhs)

        lhs = cls.ensure_if_combinable(lhs)
        rhs = cls.ensure_if_combinable(rhs)

        values, py_values = cls.split_values(lhs, rhs)

        if cls.should_optimize_for("compare", lhs, rhs):

            if values and not py_values:

                if values[0].value is None and values[1].value is None:
                    return Value(False)

                if values[0].value == values[1].value:
                    return Value(False)

                else:
                    return Value(True)

            elif py_values and not values:
                return ~Q(Exact(lhs, rhs))

        value = values[0]
        py_value = py_values[0]

        if value.value is None:
            return ~Q(IsNull(py_value, True))

        return ~Q(Exact(lhs, rhs))

    @classmethod
    def eq(
        cls,
        lhs: Any,
        rhs: Any,
    ) -> Any:
        if not cls.any_combinable(lhs, rhs):
            return eq(lhs, rhs)

        lhs = cls.ensure_if_combinable(lhs)
        rhs = cls.ensure_if_combinable(rhs)

        values, py_values = cls.split_values(lhs, rhs)

        if cls.should_optimize_for("compare", lhs, rhs):

            if values and not py_values:

                if values[0].value is None and values[1].value is None:
                    return Value(True)

                if values[0].value == values[1].value:
                    return Value(True)
                else:
                    return Value(False)

            elif py_values and not values:
                return Q(Exact(lhs, rhs))

        value = values[0]
        py_value = py_values[0]

        if value.value is None:
            return Q(IsNull(py_value, True))

        return Q(Exact(lhs, rhs))



from contextlib import nullcontext
from operator import call
from random import randint
from typing import Any, ContextManager

from django.db.models import F, Q, Value
from django.db.models.functions import Floor
from django.db.models.lookups import (
    Exact,
    GreaterThan,
    GreaterThanOrEqual,
    In, IsNull,
    LessThan,
    LessThanOrEqual,
)
from django.test import TestCase

from django_xformula import BidirectionalOperator, QueryEvaluator

__all__ = [
    "DjangoQueryEvaluatorTestCase",
]


class DjangoQueryEvaluatorTestCase(TestCase):

    evaluator: QueryEvaluator

    def setUp(self) -> None:
        self.evaluator = QueryEvaluator()

    def evaluate(
        self,
        source: str,
        context: QueryEvaluator.Context | None = None,
    ) -> Any:
        return self.evaluator.evaluate(
            source,
            context=context,
        )

    def test__evaluate__literal__none(self) -> None:
        value = self.evaluate("none")
        self.assertIs(value, None)

    def test__evaluate__literal__bool__false(self) -> None:
        value = self.evaluate("false")
        self.assertIs(value, False)

    def test__evaluate__literal__bool__true(self) -> None:
        value = self.evaluate("true")
        self.assertIs(value, True)

    def test__evaluate__literal__int__decimal(self) -> None:
        value = self.evaluate("1")
        self.assertEqual(value, 1)

    def test__evaluate__literal__float(self) -> None:
        value = self.evaluate("1.2e+4")
        self.assertEqual(value, 1.2e4)

    def test__evaluate__literal__complex(self) -> None:
        value = self.evaluate("1.2e+4j")
        self.assertEqual(value, 1.2e4j)

    def test__evaluate__literal__str__single_line__single_quoted(self) -> None:
        value = self.evaluate("'value'")
        self.assertEqual(value, "value")

    def test__evaluate__literal__str__single_line__double_quoted(self) -> None:
        value = self.evaluate('"value"')
        self.assertEqual(value, "value")

    def test__evaluate__identifier(self) -> None:
        f = self.evaluate("value")
        self.assertIsInstance(f, F)
        self.assertEqual(f.name, "value")

    def test__evaluate__identifier__with_builtin_variable(self) -> None:
        auth_user_id = 1
        context = self.evaluator.__class__.Context(
            builtins=dict(
                AUTH_USER_ID=auth_user_id,
            ),
        )
        value = self.evaluate("AUTH_USER_ID", context=context)
        self.assertEqual(value, auth_user_id)

    def test__evaluate__list__literals(self) -> None:
        container = self.evaluate("[none, true, 1, 1.2, 'abc']")
        self.assertEqual(container, [None, True, 1, 1.2, "abc"])

    def test__evaluate__list__mixed_types(self) -> None:
        context = self.evaluator.__class__.Context(
            builtins=dict(
                is_authenticated=False,
            ),
        )
        container = self.evaluate("[False, id, is_authenticated]", context=context)
        self.assertEqual(container, [False, F("id"), False])

    def test__evaluate__set(self) -> None:
        container = self.evaluate("{id, id}")
        self.assertEqual(container, {F("id")})

    def test__evaluate__tuple(self) -> None:
        container = self.evaluate("(NONE, FALSE, TRUE)")
        self.assertEqual(container, (None, False, True))

    def test__evaluate__dict(self) -> None:
        mapping = self.evaluate(
            """
            {
                a: "a",
                "b": b,
            }
            """
        )
        self.assertEqual(
            mapping,
            {
                F("a"): "a",
                "b": F("b"),
            },
        )

    def test__evaluate__function_call__void_args(self) -> None:
        context = self.evaluator.__class__.Context(
            builtins=dict(
                f=lambda: 1,
            ),
            call=call,
        )
        value = self.evaluate("f()", context=context)
        self.assertEqual(value, 1)

    def test__evaluate__function_call__args(self) -> None:
        context = self.evaluator.__class__.Context(
            builtins=dict(
                f=lambda x: x + 1,
            ),
            call=call,
        )
        value = self.evaluate("f(1)", context=context)
        self.assertEqual(value, 2)

    def test__evaluate__function_call__kwargs(self) -> None:
        def get_value(x: int = 0) -> int:
            return x

        context = self.evaluator.__class__.Context(
            builtins=dict(
                get_value=get_value,
            ),
            call=call,
        )

        value = self.evaluate("get_value()", context=context)
        self.assertEqual(value, 0)

        value = self.evaluate("get_value(x = 1)", context=context)
        self.assertEqual(value, 1)

    def test__evaluate__function_call__args_kwargs(self) -> None:
        def get_value(
            *args: int,
            **kwargs: int,
        ) -> int | None:
            if "y" in kwargs:
                return kwargs["y"]
            if not args:
                return None
            return args[0]

        context = self.evaluator.__class__.Context(
            builtins=dict(
                get_value=get_value,
            ),
            call=call,
        )

        value = self.evaluate("get_value()", context=context)
        self.assertEqual(value, None)

        value = self.evaluate("get_value(1)", context=context)
        self.assertEqual(value, 1)

        value = self.evaluate("get_value(1, y = 2)", context=context)
        self.assertEqual(value, 2)

    def test__evaluate__operation__arity_1(self) -> None:
        for (
            symbol,
            operate,
        ) in self.evaluator.__class__.Operator.Unary.dispatcher.items():

            py_symbol = self.evaluator.__class__.Operator.Unary.aliases.get(
                symbol,
                symbol,
            )

            for _ in range(100):
                value = randint(0, 10)

                py_source = f"{py_symbol} {value}"
                xf_source = f"{symbol} {value}"

                expected = eval(py_source)
                result = self.evaluate(xf_source)

                self.assertEqual(
                    (py_source, xf_source, type(result), result),
                    (py_source, xf_source, type(expected), expected),
                )

    def test__evaluate__operation__not(self) -> None:
        context = self.evaluator.__class__.Context(
            builtins=dict(
                Q=Q,
                V=Value,
            ),
            call=call,
        )

        cases = (
            ("!7", not 7),
            ("not 7", not 7),
            ("!V(0)", Value(True)),
            ("!V(1)", Value(False)),
            ("not V(1)", Value(False)),
            ("!Q()", BidirectionalOperator.QEnum.NEGATED),
            ("not Q(a=1)", ~Q(a=1)),
            ("!field", ~BidirectionalOperator.q_test(F("field"))),
        )

        for case in cases:
            self.assertEqual(
                (case[0], self.evaluate(case[0], context=context)),
                (case[0], case[1]),
            )

    def test__evaluate__operation__arity_2(self) -> None:
        excluded_symbols = {"@"}

        for (
            symbol,
            operate,
        ) in self.evaluator.__class__.Operator.Binary.dispatcher.items():

            if symbol in excluded_symbols:
                continue

            py_symbol = self.evaluator.__class__.Operator.Binary.aliases.get(
                symbol,
                symbol,
            )

            for _ in range(10):
                lhs = randint(0, 10)
                rhs = randint(0, 10)

                py_source = f"{lhs} {py_symbol} {rhs}"
                xf_source = f"{lhs} {symbol} {rhs}"

                cm: ContextManager = nullcontext()
                if symbol in ["/", "//", "%"] and rhs == 0:
                    cm = self.assertRaises(ZeroDivisionError)

                elif symbol in ["in", "!in", "not in"]:
                    py_source = f"{lhs} {py_symbol} [{rhs}]"
                    xf_source = f"{lhs} {symbol} [{rhs}]"

                # Bypass SyntaxWarning: "is / is not" with literals.
                elif symbol in ["is", "!is", "is not"]:
                    py_source = f"((lambda: {lhs})()) {py_symbol} ((lambda: {rhs})())"

                with cm:
                    expected = eval(py_source)
                    result = self.evaluate(xf_source)

                    self.assertEqual(
                        (py_source, xf_source, type(result), result),
                        (py_source, xf_source, type(expected), expected),
                    )

    def test__evaluate__operation__bidirectional__floordiv(self) -> None:
        cases = (
            ("7 // 2", 7 // 2),
            ("field // 2", Floor(F("field") / 2)),
        )

        for case in cases:
            self.assertEqual(
                (case[0], self.evaluate(case[0])),
                (case[0], case[1]),
            )

    def test__evaluate__operation__bidirectional__lshift(self) -> None:
        cases = (
            ("7 << 2", 7 << 2),
            ("field << 2", F("field").bitleftshift(2)),
        )

        for case in cases:
            self.assertEqual(
                (case[0], self.evaluate(case[0])),
                (case[0], case[1]),
            )

    def test__evaluate__operation__bidirectional__rshift(self) -> None:
        cases = (
            ("7 >> 2", 7 >> 2),
            ("field >> 2", F("field").bitrightshift(2)),
        )

        for case in cases:
            self.assertEqual(
                (case[0], self.evaluate(case[0])),
                (case[0], case[1]),
            )

    def test__evaluate__operation__bidirectional__and(self) -> None:
        context = self.evaluator.__class__.Context(
            builtins=dict(
                Q=Q,
            ),
            call=call,
        )

        cases = (
            ("7 & 2", 7 & 2),
            ("field & 2", F("field").bitand(2)),
            ("Q(a=1) & field", Q(a=1) & BidirectionalOperator.q_test(F("field"))),
            ("none & field", Value(None).bitand(F("field"))),  # type: ignore[arg-type]
            ("field & none", F("field").bitand(Value(None))),  # type: ignore[arg-type]
            ("Q(field & another) & none", ~Q(F("field").bitand(F("another")))),  # type: ignore[arg-type]
            ("Q(field & another) & false", ~Q(F("field").bitand(F("another")))),  # type: ignore[arg-type]
            ("Q(field & another) & true", Q(F("field").bitand(F("another")))),  # type: ignore[arg-type]
        )

        for case in cases:
            self.assertEqual(
                (case[0], self.evaluate(case[0], context=context)),
                (case[0], case[1]),
            )

    def test__evaluate__operation__bidirectional__xor(self) -> None:
        context = self.evaluator.__class__.Context(
            builtins=dict(
                Q=Q,
            ),
            call=call,
        )

        cases = (
            ("7 ^ 2", 7 ^ 2),
            ("field ^ 2", F("field").bitxor(2)),
            ("Q(a=1) ^ field", Q(a=1) ^ BidirectionalOperator.q_test(F("field"))),
            ("none ^ field", Value(None).bitxor(F("field"))),  # type: ignore[arg-type]
            ("field ^ none", F("field").bitxor(Value(None))),  # type: ignore[arg-type]
            ("Q(field ^ another) ^ none", Q(F("field").bitxor(F("another")))),  # type: ignore[arg-type]
            ("Q(field ^ another) ^ false", Q(F("field").bitxor(F("another")))),  # type: ignore[arg-type]
            ("Q(field ^ another) ^ true", Q(F("field").bitxor(F("another")))),  # type: ignore[arg-type]
        )

        for case in cases:
            self.assertEqual(
                (case[0], self.evaluate(case[0], context=context)),
                (case[0], case[1]),
            )

    def test__evaluate__operation__bidirectional__or(self) -> None:
        context = self.evaluator.__class__.Context(
            builtins=dict(
                Q=Q,
            ),
            call=call,
        )

        cases = (
            ("7 | 2", 7 | 2),
            ("field | 2", F("field").bitor(2)),
            ("Q(a=1) | field", Q(a=1) | BidirectionalOperator.q_test(F("field"))),
            ("none | field", Value(None).bitor(F("field"))),  # type: ignore[arg-type]
            ("field | none", F("field").bitor(Value(None))),  # type: ignore[arg-type]
            ("Q(field | another) | none", Q(F("field").bitor(F("another")))),  # type: ignore[arg-type]
            ("Q(field | another) | false", Q(F("field").bitor(F("another")))),  # type: ignore[arg-type]
            ("Q(field | another) | true", Q(F("field").bitor(F("another")))),  # type: ignore[arg-type]
        )

        for case in cases:
            self.assertEqual(
                (case[0], self.evaluate(case[0], context=context)),
                (case[0], case[1]),
            )

    def test__evaluate__operation__bidirectional__is_not(self) -> None:
        context = self.evaluator.__class__.Context(
            builtins=dict(
                Q=Q,
                V=Value,
            ),
            call=call,
        )

        cases = (
            ("7 is not 2", 7 != 2),
            ("1 is not V(2)", Value(True)),
            ("V(1) is not V(2)", Value(True)),
            ("V(1) is not V(1)", Value(False)),
            ("none !is field", ~Q(IsNull(F("field"), True))),
            ("field !is None", ~Q(IsNull(F("field"), True))),
            ("field !is 2", ~Q(Exact(F("field"), Value(2)))),
            ("field !is another", ~Q(Exact(F("field"), F("another")))),
        )

        for case in cases:
            self.assertEqual(
                (case[0], self.evaluate(case[0], context=context)),
                (case[0], case[1]),
            )

    def test__evaluate__operation__bidirectional__is(self) -> None:
        context = self.evaluator.__class__.Context(
            builtins=dict(
                Q=Q,
                V=Value,
            ),
            call=call,
        )

        cases = (
            ("7 is 2", 7 == 2),
            ("0 is V(0)", Value(True)),
            ("V(1) is V(1)", Value(True)),
            ("V(1) is V(2)", Value(False)),
            ("none is field", Q(IsNull(F("field"), True))),
            ("field is None", Q(IsNull(F("field"), True))),
            ("field is 2", Q(Exact(F("field"), Value(2)))),
            ("field is another", Q(Exact(F("field"), F("another")))),
        )

        for case in cases:
            self.assertEqual(
                (case[0], self.evaluate(case[0], context=context)),
                (case[0], case[1]),
            )

    def test__evaluate__operation__bidirectional__nlte(self) -> None:
        context = self.evaluator.__class__.Context(
            builtins=dict(
                V=Value,
            ),
            call=call,
        )

        cases = (
            ("7 !<= 2", 7 > 2),
            ("0 !<= V(0)", Value(False)),
            ("V(1) !<= V(1)", Value(False)),
            ("V(2) !<= V(1)", Value(True)),
            ("none !<= field", Value(False)),
            ("field !<= None", Value(False)),
            ("field !<= 2", Q(GreaterThan(F("field"), Value(2)))),
            ("field !<= another", Q(GreaterThan(F("field"), F("another")))),
        )

        for case in cases:
            self.assertEqual(
                (case[0], self.evaluate(case[0], context=context)),
                (case[0], case[1]),
            )

    def test__evaluate__operation__bidirectional__nlt(self) -> None:
        context = self.evaluator.__class__.Context(
            builtins=dict(
                V=Value,
            ),
            call=call,
        )

        cases = (
            ("7 !< 2", 7 >= 2),
            ("0 !< V(0)", Value(True)),
            ("V(1) !< V(1)", Value(True)),
            ("V(1) !< V(2)", Value(False)),
            ("none !< field", Value(False)),
            ("field !< None", Value(False)),
            ("field !< 2", Q(GreaterThanOrEqual(F("field"), Value(2)))),
            ("field !< another", Q(GreaterThanOrEqual(F("field"), F("another")))),
        )

        for case in cases:
            self.assertEqual(
                (case[0], self.evaluate(case[0], context=context)),
                (case[0], case[1]),
            )

    def test__evaluate__operation__bidirectional__le(self) -> None:
        context = self.evaluator.__class__.Context(
            builtins=dict(
                V=Value,
            ),
            call=call,
        )

        cases = (
            ("7 <= 2", 7 <= 2),
            ("0 <= V(0)", Value(True)),
            ("V(1) <= V(1)", Value(True)),
            ("V(2) <= V(1)", Value(False)),
            ("none <= field", Value(False)),
            ("field <= None", Value(False)),
            ("field <= 2", Q(LessThanOrEqual(F("field"), Value(2)))),
            ("field <= another", Q(LessThanOrEqual(F("field"), F("another")))),
        )

        for case in cases:
            self.assertEqual(
                (case[0], self.evaluate(case[0], context=context)),
                (case[0], case[1]),
            )

    def test__evaluate__operation__bidirectional__lt(self) -> None:
        context = self.evaluator.__class__.Context(
            builtins=dict(
                V=Value,
            ),
            call=call,
        )

        cases = (
            ("7 < 2", 7 <= 2),
            ("0 < V(0)", Value(False)),
            ("V(1) < V(1)", Value(False)),
            ("V(1) < V(2)", Value(True)),
            ("none < field", Value(False)),
            ("field < None", Value(False)),
            ("field < 2", Q(LessThan(F("field"), Value(2)))),
            ("field < another", Q(LessThan(F("field"), F("another")))),
        )

        for case in cases:
            self.assertEqual(
                (case[0], self.evaluate(case[0], context=context)),
                (case[0], case[1]),
            )

    def test__evaluate__operation__bidirectional__in(self) -> None:
        cases = (
            ("field in 2", Q(In(F("field"), Value(2)))),
            ("field in [1, 2]", Q(In(F("field"), [Value(1), Value(2)]))),
            ("field in (1, 2)", Q(In(F("field"), (Value(1), Value(2))))),
            ("field in {1, 2}", Q(In(F("field"), {Value(1), Value(2)}))),
        )

        for case in cases:
            self.assertEqual(
                (case[0], self.evaluate(case[0])),
                (case[0], case[1]),
            )

    def test__evaluate__operation__bidirectional__nin(self) -> None:
        cases = (
            ("field !in 2", ~Q(In(F("field"), Value(2)))),
            ("field not in [1, 2]", ~Q(In(F("field"), [Value(1), Value(2)]))),
            ("field !in (1, 2)", ~Q(In(F("field"), (Value(1), Value(2))))),
            ("field not in {1, 2}", ~Q(In(F("field"), {Value(1), Value(2)}))),
        )

        for case in cases:
            self.assertEqual(
                (case[0], self.evaluate(case[0])),
                (case[0], case[1]),
            )



from typing import Callable, ParamSpec, Protocol, TypeVar, runtime_checkable

__all__ = [
    "Caller",
]


P = ParamSpec("P")

R = TypeVar("R")


@runtime_checkable
class Caller(Protocol):
    def __call__(
        self,
        callee: Callable[P, R],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> R:
        ...



from django.apps import AppConfig

__all__ = [
    "DjangoXFormulaConfig",
]


class DjangoXFormulaConfig(AppConfig):

    label = "django_xformula"

    name = "django_xformula"

    verbose_name = "XFormula"



from typing import Any

from django.db.backends.base.base import BaseDatabaseWrapper
from django.db.models import Lookup
from django.db.models.sql.compiler import SQLCompiler

__all__ = [
    "Pure",
]


class Pure(Lookup):

    prepare_rhs = False

    can_use_none_as_rhs = True

    def __init__(
        self,
        expression: Any,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        super(Pure, self).__init__(expression, None)

    def as_sql(
        self,
        compiler: SQLCompiler,
        connection: BaseDatabaseWrapper,
    ) -> tuple[str, Any]:
        lhs_sql, params = self.process_lhs(compiler, connection)
        return "%s" % lhs_sql, params



from typing import Any, Protocol, runtime_checkable

__all__ = [
    "AttributeGetter",
]


@runtime_checkable
class AttributeGetter(Protocol):
    def __call__(
        self,
        owner: Any,
        atname: str,
    ) -> Any:
        ...



from django_xformula.__version__ import __version__
from django_xformula.apps import DjangoXFormulaConfig
from django_xformula.errors import ForbiddenAttribute, ForbiddenCall
from django_xformula.evaluator import BidirectionalOperator, QueryEvaluator
from django_xformula.protocols import AttributeGetter, Caller

__all__ = [
    "AttributeGetter",
    "BidirectionalOperator",
    "Caller",
    "DjangoXFormulaConfig",
    "ForbiddenAttribute",
    "ForbiddenCall",
    "QueryEvaluator",
    "__version__",
]



from django_xformula.evaluator.bidirectional_operator import BidirectionalOperator
from django_xformula.evaluator.query_evaluator import QueryEvaluator

__all__ = [
    "BidirectionalOperator",
    "QueryEvaluator",
]



from typing import Any, Callable

from django.conf import settings
from django.core.exceptions import PermissionDenied

__all__ = [
    "ForbiddenCall",
]


class ForbiddenCall(PermissionDenied):

    callee: Callable

    args: tuple[Any, ...]

    kwargs: dict[str, Any]

    def __init__(
        self,
        callee: Callable,
        args: tuple[Any, ...],
        kwargs: dict[str, Any],
        message: str | None = None,
    ) -> None:
        self.callee = callee
        self.args = args
        self.kwargs = kwargs
        if message is None:
            callee_str: str
            if not settings.DEBUG:
                callee_str = getattr(callee, "__name__", str(callee))
            else:
                callee_str = getattr(callee, "__name__", repr(callee))
            message = f"Forbidden call: {callee_str}"
        super(ForbiddenCall, self).__init__(message)



import dataclasses
from functools import partial
from operator import matmul
from typing import Any, Sequence, TypeVar, cast

from django.db.models import F, Q, Value
from xformula.syntax import Parser, ast

from django_xformula.errors import ForbiddenAttribute, ForbiddenCall
from django_xformula.evaluator.bidirectional_operator import BidirectionalOperator
from django_xformula.protocols import AttributeGetter, Caller

__all__ = [
    "QueryEvaluator",
]


T = TypeVar("T")


class QueryEvaluator:
    @dataclasses.dataclass()
    class Context:

        builtins: dict[str, Any] = dataclasses.field(
            kw_only=True,
            default_factory=dict,
        )

        call: Caller | None = dataclasses.field(
            kw_only=True,
            default=None,
        )

        getattr: AttributeGetter | None = dataclasses.field(
            kw_only=True,
            default=None,
        )

    class Operator:
        class Unary:

            aliases = {
                "!": "not",
            }

            dispatcher = {
                "+": BidirectionalOperator.pos,
                "-": BidirectionalOperator.neg,
                "!": BidirectionalOperator.not_,
                "~": BidirectionalOperator.inv,
                "not": BidirectionalOperator.not_,
            }

        class Binary:

            aliases = {
                "!in": "not in",
                "!is": "is not",
                "!<=": ">",
                "!<": ">=",
                "!>=": "<",
                "!>": "<=",
                "&&": "and",
                "||": "or",
            }

            dispatcher = {
                "**": BidirectionalOperator.pow,
                "*": BidirectionalOperator.mul,
                "@": matmul,
                "/": BidirectionalOperator.truediv,
                "//": BidirectionalOperator.floordiv,
                "%": BidirectionalOperator.mod,
                "+": BidirectionalOperator.add,
                "-": BidirectionalOperator.sub,
                "<<": BidirectionalOperator.lshift,
                ">>": BidirectionalOperator.rshift,
                "&": BidirectionalOperator.iand,
                "^": BidirectionalOperator.xor,
                "|": BidirectionalOperator.ior,
                "not in": BidirectionalOperator.nin,
                "!in": BidirectionalOperator.nin,
                "in": BidirectionalOperator.in_,
                "is not": BidirectionalOperator.is_not,
                "!is": BidirectionalOperator.is_not,
                "is": BidirectionalOperator.is_,
                "!<=": BidirectionalOperator.gt,
                "!<": BidirectionalOperator.ge,
                "<=": BidirectionalOperator.le,
                "<": BidirectionalOperator.lt,
                "!>=": BidirectionalOperator.lt,
                "!>": BidirectionalOperator.le,
                ">=": BidirectionalOperator.ge,
                ">": BidirectionalOperator.gt,
                "!=": BidirectionalOperator.ne,
                "==": BidirectionalOperator.eq,
                "&&": BidirectionalOperator.and_,
                "||": BidirectionalOperator.or_,
                "and": BidirectionalOperator.and_,
                "or": BidirectionalOperator.or_,
            }

    parser: Parser

    def __init__(
        self,
        parser: Parser | None = None,
    ) -> None:
        if parser is None:
            parser = Parser()
        self.parser = parser

    def evaluate(
        self,
        source: str,
        context: Context | None = None,
    ) -> Any:
        tree = self.parser.parse(source)
        return self.evaluate_ast(tree, context=context)

    def evaluate_ast(
        self,
        tree: ast.Node,
        context: Context | None = None,
    ) -> Any:
        return self.evaluate_node(tree, context=context)

    def evaluate_node(
        self,
        node: ast.Node,
        context: Context | None = None,
    ) -> Any:
        if context is None:
            context = self.__class__.Context()

        if isinstance(node, ast.Literal):
            return self.evaluate_literal(context, node)

        if isinstance(node, ast.Identifier):
            return self.evaluate_identifier(context, node)

        if isinstance(node, ast.Attribute):
            return self.evaluate_attribute(context, node)

        if isinstance(node, ast.Call):
            return self.evaluate_call(context, node)

        if isinstance(node, ast.List):
            return self.evaluate_list(context, node)

        if isinstance(node, ast.Set):
            return self.evaluate_set(context, node)

        if isinstance(node, ast.Tuple):
            return self.evaluate_tuple(context, node)

        if isinstance(node, ast.Dict):
            return self.evaluate_dict(context, node)

        if isinstance(node, ast.Pair):
            return self.evaluate_pair(context, node)

        if isinstance(node, ast.Operation):
            return self.evaluate_operation(context, node)

        raise NotImplementedError(
            f"{self.__class__.__qualname__}.evaluate_node is not implemented for"
            f" node: {node!r}"
        )

    def evaluate_literal(
        self,
        context: Context,
        node: ast.Literal[T],
    ) -> T:
        return node.value

    def evaluate_identifier(
        self,
        context: Context,
        node: ast.Identifier,
    ) -> str | F | ast.Identifier:
        if node.context == ast.Context.LOAD:
            if node.name in context.builtins:
                return context.builtins[node.name]
            return F(node.name)
        elif node.context == ast.Context.ATTRIBUTE:
            return node.name
        elif node.context == ast.Context.KEYWORD_ARGUMENT:
            return node.name
        raise NotImplementedError(
            f"{self.__class__.__qualname__}.evaluate_identifier"
            f" is not implemented for context: {context!r}"
        )

    def evaluate_attribute(
        self,
        context: Context,
        node: ast.Attribute,
    ) -> str | F | ast.Attribute:
        if node.context == ast.Context.LOAD:
            owner = self.evaluate_node(node.owner, context)
            attname = cast(str, self.evaluate_node(node.name, context))
            if isinstance(owner, F):
                return F(f"{owner.name}__{attname}")
            if callable(context.getattr):
                return context.getattr(owner, attname)
            raise ForbiddenAttribute(owner, attname)
        raise NotImplementedError(
            f"{self.__class__.__qualname__}.evaluate_attribute"
            f" is not implemented for context: {context!r}"
        )

    def evaluate_call(
        self,
        context: Context,
        node: ast.Call,
    ) -> Any:
        args: list[Any] = []
        kwargs: dict[str, Any] = dict()

        for element in node.arguments:
            if isinstance(element, ast.Pair):
                key = self.evaluate_node(
                    element.elements[0],
                    context=context,
                )
                key_value = self.evaluate_node(
                    element.elements[1],
                    context=context,
                )
                kwargs[key] = key_value

            else:
                value = self.evaluate_node(element, context=context)
                args.append(value)

        f = self.evaluate_node(node.callee, context=context)

        if callable(context.call):
            return context.call(
                f,
                *args,
                **kwargs,
            )

        raise ForbiddenCall(f, tuple(args), kwargs)

    def evaluate_list(
        self,
        context: Context,
        node: ast.List,
    ) -> list[Any]:
        return list(
            map(
                partial(self.evaluate_node, context=context),
                node.elements,
            ),
        )

    def evaluate_set(
        self,
        context: Context,
        node: ast.Set,
    ) -> set[Any]:
        return set(
            map(
                partial(self.evaluate_node, context=context),
                node.elements,
            ),
        )

    def evaluate_tuple(
        self,
        context: Context,
        node: ast.Tuple,
    ) -> tuple[Any, ...]:
        return tuple(
            map(
                partial(self.evaluate_node, context=context),
                node.elements,
            ),
        )

    def evaluate_dict(
        self,
        context: Context,
        node: ast.Dict,
    ) -> dict[Any, Any]:
        return dict(
            map(
                partial(self.evaluate_node, context=context),
                node.elements,
            ),
        )

    def evaluate_pair(
        self,
        context: Context,
        node: ast.Pair,
    ) -> tuple[Any, Any]:
        return (
            self.evaluate_node(node.elements[0], context=context),
            self.evaluate_node(node.elements[1], context=context),
        )

    def evaluate_operation(
        self,
        context: Context,
        node: ast.Operation,
    ) -> None | bool | int | float | complex | str | Sequence | F | Value | Q:
        if node.operator.arity == 1:
            return self.evaluate_operation_arity_1(context, node)

        if node.operator.arity == 2:
            return self.evaluate_operation_arity_2(context, node)

        return self.evaluate_operation_dynamic_arity(context, node)

    def evaluate_operation_arity_1(
        self,
        context: Context,
        node: ast.Operation,
    ) -> None | bool | int | float | complex | str | Sequence | F | Value | Q:
        operate = self.__class__.Operator.Unary.dispatcher.get(
            node.operator.symbols[0].value,
            None,
        )

        if not callable(operate):
            raise NotImplementedError(
                f"{self.__class__.__qualname__}.evaluate_operation_arity_1"
                f" is not implemented for operator: {node.operator!r}"
            )

        operand = self.evaluate_node(
            node.operands[0],
            context=context,
        )

        return operate(operand)

    def evaluate_operation_arity_2(
        self,
        context: Context,
        node: ast.Operation,
    ) -> None | bool | int | float | complex | str | Sequence | F | Value | Q:
        operate = self.__class__.Operator.Binary.dispatcher.get(
            node.operator.symbols[0].value,
            None,
        )

        if not callable(operate):
            raise NotImplementedError(
                f"{self.__class__.__qualname__}.evaluate_operation_arity_2"
                f" is not implemented for operator: {node.operator!r}"
            )

        lhs = self.evaluate_node(
            node.operands[0],
            context=context,
        )

        rhs = self.evaluate_node(
            node.operands[1],
            context=context,
        )

        return operate(lhs, rhs)

    def evaluate_operation_dynamic_arity(
        self,
        context: Context,
        node: ast.Operation,
    ) -> None | bool | int | float | complex | str | Sequence | F | Value | Q:
        raise NotImplementedError(
            f"{self.__class__.__qualname__}.evaluate_operation_dynamic_arity"
            f" is not implemented for operator: {node.operator!r}"
        )



from typing import Any

from django.conf import settings
from django.core.exceptions import PermissionDenied

__all__ = [
    "ForbiddenAttribute",
]


class ForbiddenAttribute(PermissionDenied):

    owner: Any

    attname: str

    def __init__(
        self,
        owner: Any,
        attname: str,
        message: str | None = None,
    ) -> None:
        self.owner = owner
        self.attname = attname
        if message is None:
            owner_str: str
            if not settings.DEBUG:
                owner_str = str(owner)
            else:
                owner_str = f"<{owner!r}>"
            message = f"Forbidden attribute: {owner_str}.{attname!s}"
        super(ForbiddenAttribute, self).__init__(message)



from django_xformula.errors.forbidden_attribute import ForbiddenAttribute
from django_xformula.errors.forbidden_call import ForbiddenCall

__all__ = [
    "ForbiddenAttribute",
    "ForbiddenCall",
]



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['django_xformula',
 'django_xformula.apps',
 'django_xformula.db',
 'django_xformula.db.lookups',
 'django_xformula.errors',
 'django_xformula.evaluator',
 'django_xformula.protocols',
 'django_xformula.tests',
 'django_xformula.tests.evaluator']

package_data = \
{'': ['*']}

install_requires = \
['django', 'xformula>=0.1.0,<0.2.0']

setup_kwargs = {
    'name': 'django-xformula',
    'version': '0.1.1',
    'description': 'Django query evaluator, is built on top of XFormula language front-end.',
    'long_description': "# django-xformula\n\nDjango query evaluator, is built on top of XFormula language front-end.\n\n---\n\n**This project is still in development**.\n\nIf you're interested, you may check the note in\n[XFormula](https://github.com/ertgl/xformula) repository.\n\n---\n\n\n## Features:\n\n- Bidirectional operators\n- - Same syntax for both Python and Django query evaluation\n- - If an operation contains at least one `django.db.models.expressions.Combinable`\n    or `django.db.models.Q` object, it will be evaluated as `django.db.models.Q`\n    object\n- Zero built-in variable by defaults\n- - When a variable name is used but does not exist in the specified built-ins,\n    it will be evaluated as `django.db.models.F` object\n- Customizable attribute getter; manage which attributes can be used in formulas\n  (Getting an attribute of an object is prohibited by default)\n- Customizable caller; manage which functions can be called in formulas\n  (Calling is prohibited by default)\n\n\n## License\n\n[MIT](https://github.com/ertgl/xformula/blob/main/LICENSE)\n",
    'author': 'Ertuğrul Keremoğlu',
    'author_email': 'ertugkeremoglu@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.11,<4.0',
}


setup(**setup_kwargs)
